#include "stdafx.h"
#include <string.h>
#include <stdio.h>

/*********************************************************************
 .BMP Bitmap Loading Routines
*********************************************************************/

//different header signatures for various bitmap file types
#define BFT_ICON   0x4349   /* 'IC' */
#define BFT_BITMAP 0x4d42   /* 'BM' */
#define BFT_CURSOR 0x5450   /* 'PT' */

//macro to determine to round off the given value to the closest byte
//taken from ShowDib.h, from Microsoft example source code
#define WIDTHBYTES(i)   ((i+31)/32*4)

//this function was based on example Microsoft source code from the 'ShowDib' example program
char * ReadBitmap( CString strFilename, UINT& uWidth, UINT& uHeight, HPALETTE& hPalette )
{
    CFile File;

    if( !File.Open( strFilename, CFile::modeRead|CFile::shareDenyWrite ) )
    {
        AfxMessageBox( "Cannot open bitmap file!", MB_ICONEXCLAMATION );
        return NULL;
    }

    //read the bitmap file header
    DWORD dwOffset = File.GetPosition();
    BITMAPFILEHEADER bf;

    if( File.Read( &bf, sizeof(BITMAPFILEHEADER) ) < sizeof(BITMAPFILEHEADER) )
    {
        AfxMessageBox( "Unexpected end of bitmap file!", MB_ICONEXCLAMATION );
        return NULL;
    }

    //read bitmap info header
    if( bf.bfType != BFT_BITMAP )
    {
        bf.bfOffBits = 0L;
        File.Seek( dwOffset, CFile::begin );
    }

    BITMAPINFOHEADER bi;
    if( File.Read( &bi, sizeof(BITMAPINFOHEADER) ) < sizeof(BITMAPINFOHEADER) )
    {
        AfxMessageBox( "Unexpected end of bitmap file!", MB_ICONEXCLAMATION );
        return NULL;
    }


    //determine header type
    BOOL bCoreHeader = (bi.biSize == sizeof(BITMAPCOREHEADER));

    //calculate the number of colours
    WORD wNumColours;

    if( bCoreHeader ) //image has a bitmap core header
    {
        LPBITMAPCOREHEADER lpbi = (LPBITMAPCOREHEADER)&bi;
        switch( lpbi->bcBitCount )
        {
            case 1: wNumColours = 2; break;
            case 4: wNumColours = 16; break;
            case 8: wNumColours = 256; break;
            default: wNumColours = 0; //>8 bit images have no colour table
        }
    }
    else //image has a bitmap info header
    {
        if( bi.biClrUsed != 0 )
            wNumColours = (WORD)bi.biClrUsed;
        else
        {
            switch( bi.biBitCount )
            {
                case 1: wNumColours = 2; break;
                case 4: wNumColours = 16; break;
                case 8: wNumColours = 256; break;
                default: wNumColours = 0; //>8 bit images have no colour table
            }
        }
    }


    //if the image is using a bitmap core info header, translate it to a bitmap info header
    if( bCoreHeader )
    {
        DWORD dwWidth, dwHeight;
        WORD wPlanes, wBitCount;

        //extract data from core header
        LPBITMAPCOREHEADER lpbc;
        lpbc = (LPBITMAPCOREHEADER)&bi;
        dwWidth = lpbc->bcWidth;
        dwHeight = lpbc->bcHeight;
        wPlanes = lpbc->bcPlanes;
        wBitCount = lpbc->bcBitCount;

        //insert data into info header
        bi.biSize = sizeof(BITMAPINFOHEADER);
        bi.biWidth = dwWidth;
        bi.biHeight = dwHeight;
        bi.biPlanes = wPlanes;
        bi.biBitCount = wBitCount;
        bi.biCompression = BI_RGB;
        bi.biSizeImage = 0;
        bi.biXPelsPerMeter = 0;
        bi.biYPelsPerMeter = 0;
        bi.biClrUsed = wNumColours;
        bi.biClrImportant = wNumColours;

        //adjust file pointer accordingly
        File.Seek( sizeof(BITMAPCOREHEADER) - sizeof(BITMAPINFOHEADER), CFile::current );
    }


    //check that the header size is OK now
    if( bi.biSize != sizeof(BITMAPINFOHEADER) )
    {
        AfxMessageBox( "This is not a DIB bitmap file!", MB_ICONEXCLAMATION );
        return NULL;
    }

    //fill in some default values if they are zero
    if( bi.biSizeImage == 0 )
        bi.biSizeImage = WIDTHBYTES( (bi.biWidth * bi.biBitCount) * bi.biHeight );

    //allocate a bitmap info structure
    HANDLE hbi;

    hbi = GlobalAlloc( GHND, bi.biSize + (wNumColours * sizeof(RGBQUAD)) + bi.biSizeImage );
    if( !hbi )
    {
        AfxMessageBox( "Cannot allocate memory for bitmap!", MB_ICONEXCLAMATION );
        return NULL;
    }

    LPBITMAPINFOHEADER lpbi = (LPBITMAPINFOHEADER)GlobalLock(hbi);
    *lpbi = bi;

    //get a pointer to the palette
    RGBQUAD FAR* pRGB;
    pRGB = (RGBQUAD FAR*)((LPSTR)lpbi + bi.biSize );

    //read the palette
    if( wNumColours )
    {
        DWORD dwRead;
        if( bCoreHeader )
        {
            //read the bitmap core header style palette
            dwRead = File.Read( (LPVOID)pRGB, wNumColours * sizeof(RGBTRIPLE) );
            if( dwRead != wNumColours * sizeof(RGBTRIPLE) ) 
                dwRead = 0;
            else
                //convert old (3 byte RGBTRIPLE) palette into new RGBQUAD format
                for( int iColour = wNumColours -1; iColour >= 0; iColour-- )
                {
                    RGBQUAD rgb;

                    rgb.rgbRed =   ((RGBTRIPLE FAR *)pRGB)[iColour].rgbtRed;
                    rgb.rgbGreen = ((RGBTRIPLE FAR *)pRGB)[iColour].rgbtGreen;
                    rgb.rgbBlue =  ((RGBTRIPLE FAR *)pRGB)[iColour].rgbtBlue;
                    rgb.rgbReserved = 0;

                    pRGB[iColour] = rgb;
                }
        }
        else
        {
            //read bitmap info header style palette
            dwRead = File.Read( (LPVOID)pRGB, wNumColours * sizeof(RGBQUAD) );
            if( dwRead != wNumColours * sizeof(RGBQUAD) ) dwRead = 0;
        }

        //check that all was read OK
        if( dwRead == 0 )
        {
            AfxMessageBox( "Unexpected end of file when loading palette", MB_ICONEXCLAMATION );
            GlobalUnlock( hbi );
            GlobalFree( hbi );
            return NULL;
        }
    }

    //now load the bitmap bits into memory
    DWORD dwLen = bi.biSize + (wNumColours * sizeof(RGBQUAD)) + bi.biSizeImage;

    //read in the bitmap bits
    if( bf.bfOffBits != 0L ) dwOffset = File.Seek( dwOffset + bf.bfOffBits, CFile::begin );
    if( File.Read( (LPSTR)lpbi + lpbi->biSize + (wNumColours * sizeof(RGBQUAD)), bi.biSizeImage ) != bi.biSizeImage )
    {
        AfxMessageBox( "Unexpected end of file when loading bitmap bits", MB_ICONEXCLAMATION );
        GlobalUnlock( hbi );
        GlobalFree( hbi );
        return NULL;
    }
   
    //set up the palette
    HPALETTE hPalCurrent = NULL;

    if( wNumColours )
    {
        LOGPALETTE *pPal;

        //allocate a logical palette structure
        pPal = (LOGPALETTE*)LocalAlloc( LPTR, sizeof(LOGPALETTE) + wNumColours * sizeof(PALETTEENTRY) );
        if( !pPal )
        {
            AfxMessageBox( "Unable to allocate palette", MB_ICONEXCLAMATION );
            GlobalUnlock( hbi );
            GlobalFree( hbi );
            return NULL;
        }

        pPal->palNumEntries = wNumColours;
        pPal->palVersion = 0x300;

        //copy the palette entries from the bitmap's palette
        for( int iColour = 0; iColour < wNumColours; iColour++ )
        {
            pPal->palPalEntry[iColour].peRed = pRGB[iColour].rgbRed;
            pPal->palPalEntry[iColour].peGreen = pRGB[iColour].rgbGreen;
            pPal->palPalEntry[iColour].peBlue = pRGB[iColour].rgbBlue;
            pPal->palPalEntry[iColour].peFlags = 0;
        }

        //generate the palette
        hPalCurrent = CreatePalette( pPal );
        LocalFree((HANDLE)pPal);

        //check that the palette was created OK
        if( !hPalCurrent )
        {
            AfxMessageBox( "Palette creation failed!", MB_ICONEXCLAMATION );
            GlobalUnlock( hbi );
            GlobalFree( hbi );
            return NULL;
        }

    }
    else
    {
        if( lpbi->biBitCount == 24 )
        {
            hPalCurrent = NULL;
        }
        else //unknown BPP
        {
            AfxMessageBox( "The bitmap's number of bits per pixel (BPP) is not supported", MB_ICONEXCLAMATION );
            GlobalUnlock( hbi );
            GlobalFree( hbi );
            return NULL;
        }
    }

    

    //store the width, height and number of colours
    uWidth = bi.biWidth;
    uHeight = bi.biHeight;

    hPalette = hPalCurrent;
    
    //(finally) return the handle to the bitmap info
    return (char*)lpbi;
}

/********************************************************************/


//add other image types here...

